/**************************************************************************

    display.C


    Copyright (C) 1995 Andrew Veliath

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


***************************************************************************/

// $Id: display.C,v 0.18 1996/02/17 06:59:39 drewvel Exp $             

/*******************************************************************/
#include <ostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "display.h"
#include "keymap.h"
#include "gconfig.h"
#include "defvals.h"


/*******************************************************************/
TDisplayInfo::TDisplayInfo()
{
  width=height=0;
  bg=0;
  fg=0;
  linecolor=axescolor=lowsegcolor=medsegcolor=hisegcolor=peakcolor=0;
}     


/*******************************************************************/
TDisplayControls::TDisplayControls(const TDisplay& Display)
  :display(Display)
{
  DEBUG("TDisplayControls constructing");
  
  rate.name("sample rate");
  rate.val=22050;

  FreqRange.name("freq. range");
  FreqRange.step=64;
  FreqRange.Low(64);
  FreqRange.High(256);
  
  LogX.Set();
}


/*******************************************************************/
void
TDisplayControls::CalculateFreqRangeFromZoomCenter()
{
  FreqRange.low=(int) ((float) center_freq * zoom);
  FreqRange.high=(int) ((float) (rate.val / 2 - center_freq) * (1 - zoom) + center_freq);

  if (FreqRange.low<0) FreqRange.low=0;
  if (FreqRange.high>rate.val/2) FreqRange.high=rate.val/2;

  DEBUG("New FreqRange: " << FreqRange.low << ".." << FreqRange.high);
}


/*******************************************************************/
void
TDisplayControls::CalculateZoomCenterFromFreqRange()
{
  float fh=(float) FreqRange.high, fl=(float) FreqRange.low, r=(float) rate.val/2;

  assert(r!=0);

  zoom=-(fh-fl-r)/r;

  if (zoom<0.0) zoom=0.0;
  if (zoom>1.0) zoom=1.0;

  DEBUG("Calculating zoom factor...: " << zoom);

  if (zoom<=0.0) 
    center_freq=(int) r;
  else  
    center_freq=(int) (fl/zoom);

  center_freq/=2;

  if (center_freq<0) center_freq=0;
  if (center_freq>(int) r) center_freq=(int) r;

  DEBUG("Calculating center freq...: " << center_freq);
}


/*******************************************************************/
TDisplayControls&
operator << (TDisplayControls& dc, const globalConfig& gc)
{
  DEBUG("TDisplayControls reading configuration");
  dc.ofs8=gc.Offset8.val;
  dc.ofs16=gc.Offset16.val;
  DEBUG("ofs8: " << (int) dc.ofs8 << " ofs16: " << dc.ofs16);
  dc.averaging.isOn=gc.Averaging.val;  
  dc.avgamt=gc.AvgAmt.val;
  dc.maxbands.isOn=gc.MaxBands.val;
  dc.device=gc.Device.val;
  dc.bits8.isOn=gc.Bits8.val;
  dc.rate=gc.SampleRate.val;
  dc.rate.step=256;

  dc.FreqRange.step=gc.RangeInc.val;
  dc.FreqRange.min=0;
  dc.FreqRange.max=dc.rate.val/2;
  dc.FreqRange.Low(gc.LowFreq.val);
  dc.FreqRange.High(gc.HighFreq.val);

  dc.CalculateZoomCenterFromFreqRange();
  
  if (gc.LogFreqScale.val) 
    dc.LogX.Set();
  else 
    dc.LogX.Clear();

  return dc;
}


/*******************************************************************/
const TDisplayControls&
operator >> (const TDisplayControls& dc, globalConfig& gc)
{
  DEBUG("TDisplayControls writing configuration");
  strcpy(gc.Device.val, dc.device.c_str());
  gc.Offset8.val=dc.ofs8;
  gc.Offset16.val=dc.ofs16;
  gc.Averaging.val=dc.averaging.isOn;
  gc.AvgAmt.val=dc.avgamt;
  gc.MaxBands.val=dc.maxbands.isOn;
  gc.Bits8.val=dc.bits8.isOn;
  gc.SampleRate.val=dc.rate.val;
  gc.LowFreq.val=dc.FreqRange.low;
  gc.HighFreq.val=dc.FreqRange.high;
  gc.RangeInc.val=dc.FreqRange.step;
  gc.LogFreqScale.val=dc.LogX.isOn;  

  return dc;
}



/*******************************************************************/
TDisplay::TDisplay()
  :dc(*this)
{
  DEBUG("TDisplay::TDisplay");
  running=false;
}


/*******************************************************************/
TDisplay::~TDisplay()
{
  DEBUG("TDisplay::~TDisplay");
}


/*******************************************************************/
void
TDisplay::DisplayStatus(vector<string>& ds) const
{
  DEBUG("TDisplay::DisplayStatus");
  char s[128], q[64], r[64];
  sprintf(s, 
	  "[device:%s] [%db] [bands:%d] [rate:%d] [size:%d]  ",
	  dc.device.c_str(),
	  dc.bits8.isOn ? 8 : 16,
	  Bands(),
	  dc.rate.val, 
	  WindowSize());
  ds.push_back(s);

  sprintf(r, "%-4.3f", dc.avgamt);
  if (dc.averaging.isOn)
    sprintf(q, "[avg:%s]", r);
  else
    strcpy(q, "[avg:-]");
  sprintf(s, 
	  "[low:%d] [high:%d] %-16s ",
	  dc.FreqRange.low, dc.FreqRange.high,
	  q);  
  ds.push_back(s);
}


/*******************************************************************/
void
TDisplay::MethodStatus(vector<string>&) const
{
  DEBUG("TDisplay::MethodStatus");
}


/*******************************************************************/
void
TDisplay::Initiate()
{
  DEBUG("TDisplay INITIATING");
  OnCreate();
  Resize();
  OnReset();
}


/*******************************************************************/
int
TDisplay::FFTWindowIndex(int frequency)
{
  DEBUG("FFTWindow enter");
  int WTop=WindowSize()/2;
  int idx = (int)
    ((float) WTop * (float) frequency/((float) dc.rate.val/2.0));

  if (idx<0) idx=0;
  if (idx>=WTop) idx=WTop-1;

  DEBUG("FFTWindow index gets " << frequency << " returns " << idx);
    
  return idx;
}


/*******************************************************************/
void
TDisplay::BandScale()
{
  DEBUG("band scaling start");

  int low=FFTWindowIndex(dc.FreqRange.low), high=FFTWindowIndex(dc.FreqRange.high);

  band_base.scale(low, high, dc.LogX.isOn);
  band_length.scale(low, high, dc.LogX.isOn);

  int last=band_length.length()-1, i;
  for (i=0; i<last; i++) {
    band_length.index()[i]=band_length.index()[i+1]-band_length.index()[i];
    if (band_length.index()[i]<1) band_length.index()[i]=1;
  }
  
  band_length.index()[last]=fft.length()/2-band_length.index()[last];
  if (band_length.index()[last]<1) band_length.index()[last]=1;

  frequencies.scale(0, dc.rate.val);
}


/*******************************************************************/
int
TDisplay::WindowSize() const
{
  return fft.length();
}


/*******************************************************************/
int
TDisplay::WindowSize(int w)
{
  DEBUG("TDisplay Window (" << w << ")");

  int old=fft.length(w);
  frequencies.length(w);
  frequencies.scale(0, dc.rate.val);

  return old;
}


/*******************************************************************/
int
TDisplay::Bands() const
{
  return band_base.length();
}


/*******************************************************************/
int
TDisplay::Bands(int bands)
{
  DEBUG("TDisplay Bands (" << bands << ")");

  int old=Bands();

  if (bands>1) {
    band_base.length(bands);    
    band_length.length(bands);
  }

  return old;
}


/*******************************************************************/
void
TDisplay::OnCreate()
{
  DEBUG("TDisplay OnCreate");
}


/*******************************************************************/
void
TDisplay::OnHelp()
{
  DEBUG("TDisplay OnHelp");
  OnDisplayHelp();
  OnMethodHelp();
}


/*******************************************************************/
void
TDisplay::OnReset()
{
  DEBUG("TDisplay OnReset");
}


/*******************************************************************/
void
TDisplay::OnResize()
{
  DEBUG("TDisplay resize message received");
  DEBUG("    :new size is " << di.width << "x" << di.height);
  DEBUG("Width is " << di.width);
  
  if (dc.maxbands.isOn) Bands(GraphWidth());
}


/*******************************************************************/
bool
TDisplay::OnStart()
{
  DEBUG("TDisplay OnStart");
  return true;
}


/*******************************************************************/
void
TDisplay::OnStop()
{
  DEBUG("TDisplay OnStop");
}               


/*******************************************************************/
void
TDisplay::OnUpdate()
{
  DEBUG("TDisplay OnUpdate");
}


/*******************************************************************/
bool
TDisplay::OnStartMethod()
{
  DEBUG("TDisplay [METHOD] OnStart");
  return true;
}


/*******************************************************************/
void
TDisplay::OnStopMethod()
{
  DEBUG("TDisplay [METHOD] OnStop");
}


/*******************************************************************/
void
TDisplay::OnUpdateAxes()
{
  DEBUG("TDisplay OnUpdateAxes");
}


/*******************************************************************/
void
TDisplay::OnUpdateControls()
{
  DEBUG("TDisplay OnUpdateControls");
}


/*******************************************************************/
int
TDisplay::QueryFrequencyFromPoint(int, int)
{
  return 0;
}


/*******************************************************************/
int
TDisplay::QueryAmplitudeFromPoint(int, int)
{
  return 0;
}


/*******************************************************************/
void
TDisplay::OnResizeMethod()
{
  DEBUG("TDisplay OnResizeMethod");
}


/*******************************************************************/
void
TDisplay::ClearGraph()
{
  DEBUG("TDisplay ClearGraph");
}


/*******************************************************************/
void
TDisplay::SetColor(unsigned long)
{
}


/*******************************************************************/
void
TDisplay::DrawLine(int, int, int, int)
{
}     


/*******************************************************************/
void
TDisplay::DrawLine(int, int, int, int, unsigned long)
{
}     


/*******************************************************************/
void
TDisplay::DrawLines(XPoint*, int, int)
{
}     


/*******************************************************************/
void
TDisplay::DrawText(int, int, const char*)
{
}     


/*******************************************************************/
void
TDisplay::DrawCenteredText(int, int, const char*)
{
}     


/*******************************************************************/
void
TDisplay::FillRectangle(int, int, int, int)
{
}     


/*******************************************************************/
void
TDisplay::FillRectangle(int, int, int, int, unsigned long)
{
}     


/*******************************************************************/
void
TDisplay::EnableClipping()
{
  DEBUG("TDisplay EnableClipping");
}     


/*******************************************************************/
void
TDisplay::DisableClipping()
{
  DEBUG("TDisplay DisableClipping");
}     



/*******************************************************************/
void
TDisplay::Refresh(int flags)
{
  DEBUG("TDisplay::Refresh: flags == " << flags);
  if (flags & UpdateSize) Resize();
  if (flags & UpdateAxes) {
    BandScale();
    OnUpdateAxes();    
  }
  if (flags & UpdateControls) OnUpdateControls();
}       


/*******************************************************************/
void
TDisplay::Resize()
{
  DEBUG("TDisplay::Resize");
  OnResize();
  OnResizeMethod();
  ClearGraph();
}       


/*******************************************************************/
bool
TDisplay::Start()
{
  if (!running)
    return running=OnStartMethod() && OnStart();
  else 
    return true;
}


/*******************************************************************/
void
TDisplay::Stop()
{
  if (running) {
    running=false;
    OnStopMethod();
    OnStop();
  }
}       


/*******************************************************************/
int
TDisplay::ProcessControl(char control)
{
  return 
    ProcessDisplayControl(control) | ProcessMethodControl(control);
}


/*******************************************************************/
void
TDisplay::Show()
{
  DEBUG("TDisplay Show");
}       


/*******************************************************************/
void
TDisplay::Hide()
{
  DEBUG("TDisplay Hide");
  Stop();
}       


/*******************************************************************/
void
TDisplay::OnSetConfig(const globalConfig& gc)
{
  DEBUG("TDisplay OnSetConfig");
  OnSetDisplayConfig(gc);
  OnSetMethodConfig(gc);
}


/*******************************************************************/
void
TDisplay::OnRetrieveConfig(globalConfig& gc)
{
  DEBUG("TDisplay OnRetreiveConfig");
  OnRetrieveDisplayConfig(gc);
  OnRetrieveMethodConfig(gc);  
}


/*******************************************************************/
void
TDisplay::OnSetDisplayConfig(const globalConfig& gc)
{
  DEBUG("TDisplay OnSetDisplayConfig");

  dc << gc;
  Bands(gc.NumBands.val);
  WindowSize(gc.SampleSize.val);  
}


/*******************************************************************/
void
TDisplay::OnRetrieveDisplayConfig(globalConfig& gc)
{
  DEBUG("TDisplay OnRetreiveDisplayConfig");

  dc >> gc;
  gc.NumBands.val=Bands();
  gc.SampleSize.val=WindowSize();
}


/*******************************************************************/
int
TDisplay::ProcessDisplayControl(char control)
{
  DEBUG("Processing TDisplay::DisplayControls : " << control);

  int rv=0;

  switch (control) {

  case ')':
  case '(':
    {
      int nbands=Bands();
      Bands(control==')' ? ++nbands : --nbands);
      rv|=UpdateSize | UpdateControls;
    }
    break;

  case K_TOGGLEBITRES:
    dc.bits8.Toggle();
    rv|=UpdateAxes | UpdateControls | UpdateSignal;
    break;

  case K_AVERAGING:
    dc.averaging.Toggle();
    rv|=UpdateControls;
    break;

  case K_LOWFREQ_INC:
    dc.FreqRange.LowInc();
    rv|=UpdateAxes | UpdateControls;
    break;

  case K_LOWFREQ_DEC:
    dc.FreqRange.LowDec();
    rv|=UpdateAxes | UpdateControls;
    break;

  case K_HIGHFREQ_INC:
    dc.FreqRange.HighInc();
    rv|=UpdateAxes | UpdateControls;
    break;

  case K_HIGHFREQ_DEC:
    dc.FreqRange.HighDec();
    rv|=UpdateAxes | UpdateControls;
    break;

  case K_SAMPLERATE_INC:
  case K_SAMPLERATE_DEC:
    dc.rate.val += (control==K_SAMPLERATE_INC) ? D_INT_SRATEINC : -D_INT_SRATEINC;
    rv|=UpdateAxes | UpdateControls | UpdateSignal;
    break;

  case K_CLEARGRAPH:
    ClearGraph();
    break;

  case K_LOGXSCALE_TOGGLE:
    dc.LogX.Toggle();
    rv|=UpdateControls|UpdateAxes;
    break;

  case K_HELP:
    Hide();
    OnHelp();
    break;
  }

  return rv;
}                                                            


/*******************************************************************/
void
TDisplay::OnSetMethodConfig(const globalConfig&)
{
  DEBUG("TDisplay OnSetMethodConfig");
}


/*******************************************************************/
void
TDisplay::OnRetrieveMethodConfig(globalConfig&)
{
  DEBUG("TDisplay OnRetreiveMethodConfig");
}


/*******************************************************************/
