/**************************************************************************

    display.h

    Copyright (C) 1995 Andrew Veliath

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


***************************************************************************/

// $Id: display.h,v 0.15 1996/02/17 06:54:58 drewvel Exp $             

/*******************************************************************/
#ifndef __DISPLAY_H
#define __DISPLAY_H

/*******************************************************************/
#include <string>
#include <vector>

#include "common.h"
#include "fftclass.h"
#include "index.h"
#include "controls.h"
#include "signal.h"

typedef int Dimension;
typedef struct { short x, y; } XPoint;
#define CoordModeOrigin 0 
#define CoordModePrevious 1
#define ThisPercentOf(p,a) ((int)(((float) p/100.0)*(float) a))

// update flags
#define _Bit(a) ((int)(1L<<(a)))

#define UpdateControls _Bit(0)
#define UpdateAxes     _Bit(1)
#define UpdateSize     _Bit(2)
#define UpdateSignal   _Bit(3)

struct Resources;
struct globalConfig;

/*******************************************************************/
class TDisplayInfo {

public:
  Resources* resources;

  Dimension width, height;

  unsigned long fg, bg;
  unsigned long linecolor, axescolor, labelcolor;
  unsigned long lowsegcolor, medsegcolor, hisegcolor, peakcolor;

  TDisplayInfo();
};      

class TDisplay;

/*******************************************************************/
class TDisplayControls {

friend TDisplayControls& operator << (TDisplayControls&, const globalConfig&);
friend const TDisplayControls& operator >> (const TDisplayControls&, globalConfig&);

private:
  const TDisplay& display;

  void CalculateFreqRangeFromZoomCenter();
  void CalculateZoomCenterFromFreqRange();

public:
  TDisplayControls(const TDisplay&);
  
  TSignal::sample8 ofs8;
  TSignal::sample16 ofs16;
  string device;
  IntControl rate;
  RangeControl FreqRange;
  BoolControl LogX;  
  BoolControl bits8;
  BoolControl averaging;
  BoolControl maxbands;
  real avgamt;
  int center_freq;
  float zoom;
};


/*******************************************************************/
TDisplayControls& operator << (TDisplayControls&, const globalConfig&);
const TDisplayControls& operator >> (const TDisplayControls&, globalConfig&);

/*******************************************************************/
class TDisplay {

friend class TDisplayControls;
friend class TAnalyzer;
friend class TDMStatus;
  
private:  
  bool running;

  int FFTWindowIndex(int);

protected:
  TFFT fft;
  TDisplayInfo di;
  TDisplayControls dc;

  TIndex<int> frequencies;
  TIndex<short> band_base, band_length;

  virtual void DisplayStatus(vector<string>&) const;
  virtual void MethodStatus(vector<string>&) const;

  virtual int WindowSize() const;
  virtual int WindowSize(int);
  virtual int Bands() const;
  virtual int Bands(int);
  virtual void BandScale();
  virtual int GraphWidth() const { return di.width; }
  virtual int GraphHeight() const { return di.height; }

  // messages
  virtual void OnCreate();
  virtual void OnHelp();
  virtual void OnReset();
  virtual void OnResize();
  virtual void OnResizeMethod();
  virtual bool OnStart();
  virtual bool OnStartMethod();
  virtual void OnStop();
  virtual void OnStopMethod();
  virtual void OnUpdate();
  virtual void OnUpdateAxes();
  virtual void OnUpdateControls();

  // method queries
  virtual int QueryFrequencyFromPoint(int,int);
  virtual int QueryAmplitudeFromPoint(int,int);

  // drawing functions
  virtual void ClearGraph();
  virtual void DrawLine(int, int, int, int);
  virtual void DrawLine(int, int, int, int, unsigned long);
  virtual void DrawLines(XPoint*, int, int);
  virtual void DrawText(int, int, const char *);
  virtual void DrawCenteredText(int, int, const char *);
  virtual void FillRectangle(int, int, int, int);
  virtual void FillRectangle(int, int, int, int, unsigned long);
  virtual void SetColor(unsigned long);

  virtual void EnableClipping();
  virtual void DisableClipping();

  virtual void Refresh(int flags=0);
  void Resize();
  bool Start();
  void Stop();

  // control signals 
  virtual int ProcessControl(char); 

  // these are called by ProcessControl (return flag combination of what to update)
  virtual int ProcessDisplayControl(char);
  virtual int ProcessMethodControl(char) { return 0; }

  // for display and method invocation only
  virtual void OnDisplayHelp() { }
  virtual void OnMethodHelp() { }
  virtual void OnSetDisplayConfig(const globalConfig&);
  virtual void OnRetrieveDisplayConfig(globalConfig&);
  virtual void OnSetMethodConfig(const globalConfig&);
  virtual void OnRetrieveMethodConfig(globalConfig&);

public: 
  TDisplay();
  virtual ~TDisplay();

  // main instantiation function
  virtual void Initiate();

  virtual void Show();
  virtual void Hide();

  // global configuration accessors
  virtual void OnSetConfig(const globalConfig&);
  virtual void OnRetrieveConfig(globalConfig&);
};

#endif /* __DISPLAY_H */
