/**************************************************************************

    svga_display.C

    Copyright (C) 1996 Andrew Veliath

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


***************************************************************************/

/* $Id: svga_display.C,v 0.7 1996/02/21 21:15:26 drewvel Exp $ */


/*******************************************************************/
#include <stdio.h>
#include <sys/soundcard.h>

#include "../dmstatus.h"
#include "../keymap.h"
#include "../defvals.h"

#include "svga_display.h"
#include "svga_analyzer.h"
#include "svga_color.h"

#define SCRBOXWIDTH 6
#define SCRWIDTH (WIDTH-SCRBOXWIDTH*2)
#define SCRHEIGHT (HEIGHT-SCRBOXWIDTH*2)

#define BOXWIDTH 10
#define NUMBITS analyzer->signal.bits()


/*******************************************************************/
void 
SVGA_doBox(int x1, int y1, int x2, int y2, int col)
{
  gl_line(x1, y1, x2, y1, col);
  gl_line(x1, y1, x1, y2, col);
  gl_line(x2, y1, x2, y2, col);
  gl_line(x1, y2, x2, y2, col);
}


/*******************************************************************/
void 
SVGA_FadeBox(int x1, int y1, int x2, int y2,
	     const TColor& from_col, const TColor& to_col, 
	     int width, bool fill, int bgcol)
{
  int i;
  int* cols=new int[width];
  TColor *Colors=new TColor[2];
    
  Colors[0]=from_col;
  Colors[1]=to_col;

  SVGA_FadeColorRange(Colors, 2, cols, width);
  
  for (i=0; i<width; i++) 
    SVGA_doBox(x1-i, y1-i, 
	       x2+i, y2+i, 
	       cols[i]);
  
  delete [] Colors;
  delete [] cols;
  
  if (fill) 
    gl_fillbox(x1+1, y1+1, 
	       x2-x1+1, y2-y1+1, 
	       bgcol);
  
}


/*******************************************************************/
TSVGA_DisplayInfo&
operator << (TSVGA_DisplayInfo& sdi, const globalConfig& gc)
{
  sdi.FG=gc.FG;
  sdi.ScreenBG=gc.BG;
  sdi.BG=gc.GraphBG;
  sdi.LineColor=gc.LineColor;
  sdi.AxesColor=gc.AxesColor;
  sdi.LabelColor=gc.LabelColor;
  sdi.LowSegColor=gc.LowSegColor;
  sdi.MedSegColor=gc.MedSegColor;
  sdi.HiSegColor=gc.HiSegColor;
  sdi.PeakColor=gc.PeakColor;
  return sdi;
}


/*******************************************************************/
const TSVGA_DisplayInfo&
operator >> (const TSVGA_DisplayInfo& sdi, globalConfig& gc)
{
  gc.FG=sdi.FG;
  gc.BG=sdi.ScreenBG;;
  gc.GraphBG=sdi.BG;
  gc.LineColor=sdi.LineColor;
  gc.AxesColor=sdi.AxesColor;
  gc.LabelColor=sdi.LabelColor;
  gc.LowSegColor=sdi.LowSegColor;
  gc.MedSegColor=sdi.MedSegColor;
  gc.PeakColor=sdi.PeakColor;
  return sdi;
}


/*******************************************************************/
void
TSVGA_Display::init()
{
  fp=NULL;
  screen_on=false;
  prevmode=mode=0;
  analyzer=NULL;
}


/*******************************************************************/
TSVGA_Display::TSVGA_Display(TSVGA_Analyzer* Analyzer)
{
  DEBUG("SVGA_Display constructing");
  init();
  assert(Analyzer);
  analyzer=Analyzer;
}


/*******************************************************************/
TSVGA_Display::~TSVGA_Display()
{
  DEBUG("SVGA_Display destructing");
  assert(fp);
  delete fp;

  if (prevmode)
    vga_setmode(prevmode);
}


/*******************************************************************/
void
TSVGA_Display::OnCreate()
{
  DEBUG("SVGA_Display OnCreate");
  
  assert(!screen_on);

  mode=vga_getdefaultmode();

  if (mode<10)
    if (mode==-1) 
      mode=G640x480x256;
    else { 
      cerr << "error: " NAME " requires at least G640x480x256 video mode" << endl;
      cerr << "       (GSVGAMODE default was mode " << mode << ")" << endl;
      exit(1);
    }
  
  if (vga_hasmode(mode)) {
    
    vga_setmousesupport(1);
    mouse_setwrap(MOUSE_NOWRAP);
    prevmode=vga_setmode(mode);
    screen_on=true;

    gl_setcontextvga(mode);
    
    gl_setwritemode(WRITEMODE_MASKED);
    gl_setrgbpalette();
    
    assert(!fp);
    
    if ((fp=new char[FWIDTH*FHEIGHT*256*BYTESPERPIXEL])) {
      gl_expandfont(FWIDTH, FHEIGHT, 255, gl_font8x8, fp);
      gl_setfont(FWIDTH, FHEIGHT, fp);
      DEBUG("font allocated");
    } 
    else {
      cerr << "couldn't allocate SVGAlib font!" << endl;
      exit(1); 
    }
    
    graph_x=BOXWIDTH+SCRBOXWIDTH;
    graph_y=FHEIGHT*5+BOXWIDTH+SCRBOXWIDTH;
    di.width=WIDTH-(SCRBOXWIDTH+BOXWIDTH)-graph_x;
    di.height=HEIGHT-SCRBOXWIDTH-BOXWIDTH-graph_y;

    DEBUG("graph window calculated to: x: " 
	  << graph_x << " y: " << graph_y << " w: " << di.width << " h: " << di.height);

    // setup colors
    di.fg=SVGA_RGBColor(sdi.FG);
    sdi.screenbg=SVGA_RGBColor(sdi.ScreenBG);
    di.bg=SVGA_RGBColor(sdi.BG);
    di.linecolor=SVGA_RGBColor(sdi.LineColor);
    di.axescolor=SVGA_RGBColor(sdi.AxesColor);
    di.labelcolor=SVGA_RGBColor(sdi.LabelColor);
    di.lowsegcolor=SVGA_RGBColor(sdi.LowSegColor);
    di.medsegcolor=SVGA_RGBColor(sdi.MedSegColor);
    di.hisegcolor=SVGA_RGBColor(sdi.HiSegColor);
    di.peakcolor=SVGA_RGBColor(sdi.PeakColor);
  } 
  else {
    cerr << "error: SVGAlib says your video card doesn't support mode " << mode << endl;
    exit(1);
  }
}


/*******************************************************************/
void
TSVGA_Display::OnDisplayHelp()
{
  extern char *gProgInfo;
  cerr << gProgInfo << endl;
}


/*******************************************************************/
void
TSVGA_Display::OnReset()
{
  DEBUG("TSVGA_Display OnReset");

  assert(screen_on);
  
  gl_setcontextvga(mode);
  
  gl_setwritemode(WRITEMODE_MASKED);
  gl_setrgbpalette();
  
  SVGA_FadeBox(SCRBOXWIDTH, SCRBOXWIDTH,
	       SCRWIDTH+SCRBOXWIDTH, SCRHEIGHT+SCRBOXWIDTH,
	       sdi.ScreenBG, TColor(0,0,0),
	       SCRBOXWIDTH, true, sdi.screenbg);

  ClearGraph();

  if (!analyzer->gconfig.NoText.val)
    DrawTextP(1, 1, QNAME, di.fg);

  OnUpdateAxes();

  OnUpdateControls();

  DEBUG("TSVGA_Display OnReset DONE");
}


/*******************************************************************/
void
TSVGA_Display::OnUpdateControls()
#define SwitchOn(a,b) ((a) ? (b) : ' ')
#define WordOn(a,b,c) ((a) ? (b) : (c))
#define XCharPos(x) (FWIDTH*(x-1)+1)
#define YCharPos(y) ((FHEIGHT+1)*(y-1)+1)
{
  DEBUG("SVGA: updating controls");
  assert(screen_on);
  assert(analyzer);
  
  if (!analyzer->gconfig.NoText.val) {

    // use the status gathering object to get state information
    TDMStatus s(*this); 
    vector<string>::const_iterator 
      ds=s.DisplayStats().begin(),
      ms=s.MethodStats().begin();
    
    DrawTextP(XCharPos(strlen(QNAME)+1), YCharPos(1), " -- ", di.fg, sdi.screenbg);

    int yPos=1;
    DrawTextP(XCharPos(strlen(QNAME)+5), YCharPos(yPos++), 
	      ds->c_str(), di.fg, sdi.screenbg);
    ds++;

    while (ds!=s.DisplayStats().end()) {
      DrawTextP(XCharPos(1), YCharPos(yPos++), ds->c_str(), di.fg, sdi.screenbg);
      ds++;
    }

    while (ms!=s.MethodStats().end()) {
      DrawTextP(XCharPos(1), YCharPos(yPos++), ms->c_str(), di.fg, sdi.screenbg);
      ms++;
    }
  }
}
#undef YCharPos
#undef XCharPos
#undef SwitchOn
#undef WordOn


/*******************************************************************/
void
TSVGA_Display::Show()
{
  TDisplay::Show();
  DEBUG("TSVGA_Display Show");

  if (!screen_on) {
    vga_setmode(mode);
    screen_on=true;
    OnReset();
  }  
}       


/*******************************************************************/
void
TSVGA_Display::Hide()
{
  TDisplay::Hide();
  DEBUG("TSVGA_Display Hide");
  
  if (screen_on) {
    vga_setmode(prevmode);
    screen_on=false;
  }
}       


/*******************************************************************/
void
TSVGA_Display::ClearGraph()
{
  assert(screen_on);
  
  SVGA_FadeBox(graph_x, graph_y,
	       graph_x+di.width, graph_y+di.height,
	       sdi.BG, sdi.ScreenBG, 
	       BOXWIDTH, true, di.bg);
  
  OnUpdateAxes();
}


/*******************************************************************/
void
TSVGA_Display::DrawTextP(int x, int y, const char* str, int fg, int bg)
{
  assert(fp);

  gl_setwritemode(WRITEMODE_OVERWRITE|FONT_COMPRESSED);
  gl_setfont(FWIDTH, FHEIGHT, gl_font8x8);

  if (bg==-1) 
    gl_setfontcolors(sdi.screenbg, fg);
  else 
    gl_setfontcolors(bg, fg);

#ifdef SVGALIB_CHAR_CONST_CAST
  char* nc_str=const_cast<char*>(str);
  gl_write(x+SCRBOXWIDTH, y+SCRBOXWIDTH, nc_str);
#else
  gl_write(x+SCRBOXWIDTH, y+SCRBOXWIDTH, str);
#endif

  gl_setwritemode(WRITEMODE_MASKED|FONT_EXPANDED);
  gl_setfont(FWIDTH, FHEIGHT, fp);
}     

void
TSVGA_Display::CrossY(int x, int y, int h, int state)
{
  static int ox, oy, oh;
  static char* lbuf;
  switch (state) {
  case 0:
    lbuf=new char[h*BYTESPERPIXEL*2];
    gl_getbox(graph_x+x, graph_y+y, 1, h+1, lbuf);
    DrawLine(x, y, x, y+h, di.axescolor);
    ox=x; oy=y; oh=h;
    break;
    
  case 1:
    gl_putbox(graph_x+ox, graph_y+oy, 1, oh+1, lbuf);
    gl_getbox(graph_x+x, graph_y+y, 1, h+1, lbuf);
    DrawLine(x, y, x, y+h, di.axescolor);
    ox=x; oy=y; oh=h;
    break;

  case 2:
    gl_putbox(graph_x+ox, graph_y+oy, 1, oh+1, lbuf);
    delete [] lbuf;
    break;
  }
}


/*******************************************************************/
void
TSVGA_Display::OnPause()
{
  DEBUG("TSVGA_Display::OnPause");

  gl_setwritemode(WRITEMODE_OVERWRITE);
  static int x=di.width/2, y=di.height/2, h=GraphHeight();
  int dx=0, dy=0;
  char msg[256];

  mouse_setposition(x, y);

  CrossY(x, 0, h, 0);
  
  bool done=false, update=true;
  while (!done) {

    switch (vga_getkey()) {
    case 0:
      break;
      
    case 27:
      if (vga_getkey()!=91) { vga_getkey(); break; }
      
      switch (vga_getkey()) {
      case 49: // home
	mouse_setposition(graph_x, mouse_gety());
	update=true;
	vga_getkey();
	break;

      case 52: // end
	mouse_setposition(di.width, mouse_gety());
	update=true;
	vga_getkey();
	break;

      case 67: // right
	mouse_setposition(mouse_getx()+1, mouse_gety());
	update=true;
	break;

      case 68: // left
	mouse_setposition(mouse_getx()-1, mouse_gety());
	update=true;
	break;
      }
      break;

    case ',':
      mouse_setposition(mouse_getx()-1, mouse_gety());
      update=true;
      break;

    case '.':
      mouse_setposition(mouse_getx()+1, mouse_gety());
      update=true;
      break;

    default:
      done=true;
      break;
    }
    
    if (mouse_update() || update) {
      update=false;
      int button=mouse_getbutton(), nx=mouse_getx(), ny=mouse_gety();
      if (nx<graph_x) nx=graph_x;
      if (nx>di.width) nx=di.width;
      dx=nx-x; dy=ny-y;
      x=nx; y=ny;
      CrossY(x, 0, h, 1);
      int freq=QueryFrequencyFromPoint(x, y);
      int amp=QueryAmplitudeFromPoint(x, y);
      sprintf(msg, "Frequency: %dHz  Amplitude: %d           ", freq, amp);
      DrawText(1,di.height-8,msg);
      done=button & MOUSE_LEFTBUTTON;
    }
  }
  sprintf(msg, "%40c", ' ');
  DrawText(1,di.height-8,msg);
  CrossY(x, 0, h, 2);
  gl_setwritemode(WRITEMODE_MASKED);
}


/*******************************************************************/
bool
TSVGA_Display::OnStart()
{
  bool rv=false;

  // zero averaging buffers
  analyzer->avg8=dc.ofs8; analyzer->avg16=dc.ofs16;

  DEBUG("TSVGA_Display OnStart");
  DEBUG("Device filename: " << dc.device.c_str());
  if (!analyzer->signal.open())
    if (analyzer->signal.open(dc.device.c_str())) {

      analyzer->OnUpdateSignalParameters();
#if 0
      analyzer->gconfig.SampleRate.val=dc.rate.val;

      if (analyzer->gconfig.HighFreq.val==-1)
        analyzer->gconfig.HighFreq.val=analyzer->signal.sample_rate()/2;
#endif
      
      rv=true;
    }
    else {
      DEBUG(" -- open nonsuccessful");
    }

  return rv;
}


/*******************************************************************/
void
TSVGA_Display::OnStop()
{
  DEBUG("TSVGA_Display OnStop -- CLOSING SIGNAL");
  analyzer->signal.close();
}               


/*******************************************************************/
int
TSVGA_Display::ProcessDisplayControl(char control)
{
  int rv=TDisplay::ProcessDisplayControl(control);

  DEBUG("Processing TSVGA_Display::DisplayControls : " << control);

  switch (control) {

  case K_PAUSE:
    Stop();
    OnPause();
    ClearGraph();
    Start();
    break;
  }

  return rv;
}                                                            


/*******************************************************************/
void
TSVGA_Display::OnSetDisplayConfig(const globalConfig& gc)
{
  TDisplay::OnSetDisplayConfig(gc);

  DEBUG("TSVGA_Display OnSetDisplayConfig");

  sdi << gc;  
}


/*******************************************************************/
void
TSVGA_Display::OnRetrieveDisplayConfig(globalConfig& gc)
{
  TDisplay::OnRetrieveDisplayConfig(gc);

  DEBUG("TSVGA_Display OnRetreiveDisplayConfig");

  sdi >> gc;
}


/*******************************************************************/
